#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>


#include <asm/rt2880/rt_mmap.h>
#include "rt_config.h"
#include "define.h"
#include "chip/rt2880.h" 
//#include "rt2860.h"
//#include "2570.h"
//#pragma SAVE
//#pragma	OPTIMIZE(8)


#define LED_CHECK_INTERVAL 		500 /* 50ms */
#define LED_CHECK_INTERVAL_WPS 	50 /* 1ms */

#ifdef CONFIG_RT2860V2_AP_LED_GPIO_9
#define LED_GPIO  9
#else
#define LED_GPIO 11
#endif


#define RALINK_PRGIO_ADDR		RALINK_PIO_BASE // Programmable I/O
#define RALINK_REG_PIODATA		(RALINK_PRGIO_ADDR + 0x20)

#define RALINK_GPIO_0			0x00000001
#define RALINK_GPIO_1			0x00000002
#define RALINK_GPIO_2			0x00000004
#define RALINK_GPIO_3			0x00000008
#define RALINK_GPIO_4			0x00000010
#define RALINK_GPIO_5			0x00000020
#define RALINK_GPIO_6			0x00000040
#define RALINK_GPIO_7			0x00000080
#define RALINK_GPIO_8			0x00000100
#define RALINK_GPIO_9			0x00000200
#define RALINK_GPIO_10			0x00000400
#define RALINK_GPIO_11			0x00000800
#define RALINK_GPIO_12			0x00001000
#define RALINK_GPIO_13			0x00002000
#define RALINK_GPIO_14			0x00004000
#define RALINK_GPIO_15			0x00008000
#define RALINK_GPIO_16			0x00010000
#define RALINK_GPIO_17			0x00020000
#define RALINK_GPIO_18			0x00040000
#define RALINK_GPIO_19			0x00080000
#define RALINK_GPIO_20			0x00100000
#define RALINK_GPIO_21			0x00200000
#define RALINK_GPIO_22			0x00400000
#define RALINK_GPIO_23			0x00800000
#define RALINK_GPIO(x)			(1 << x)


#define __LED_ON(gpio) ralink_gpio_led_value |= RALINK_GPIO(gpio);
#define __LED_OFF(gpio) ralink_gpio_led_value &= ~RALINK_GPIO(gpio);

u32 ralink_gpio_led_value = 0;


RALINK_TIMER_STRUCT LedCheckTimer;
RALINK_TIMER_STRUCT LedCheckTimer_wps;

void LedCtrlMain(void);
void LedCtrlMain_wps(void);

void SetLinkStatus(void);


unsigned char LinkStatus;
unsigned char SignalStrength;
unsigned char GPIOPolarity;
volatile unsigned char *McuIntStatus = (unsigned char *)(RTMP_MAC_CSR_ADDR + 0x0414);
union uLedParameter LedParameter;

extern void ralink_gpio_led_set(int gpio, unsigned int on, unsigned int off, unsigned int blinks, unsigned int rests, unsigned int times);

void LedCtrlMainEntry(void)	{	LedCtrlMain();		}	// @ 0x1022
void SetLinkStatusEntry(void)	{	SetLinkStatus();	}	// @ 0x1025
BYTE CheckTimerEbl=0;
BYTE CheckTimerEbl_wps=0;

enum LED_STATE { SOLID_OFF=0, SOLID_ON, BLINK } led_state ;

// Called when host sets LED mode.
// Called in interrupt level
void SetLinkStatus()
{	
	if(CheckTimerEbl==0){
		RTMPInitTimer(NULL, &LedCheckTimer, LedCtrlMain, NULL, TRUE);
		LedCtrlMain();
		CheckTimerEbl=1;
		led_state=SOLID_OFF ;
	}	
	//printk(KERN_INFO "%s : LinkStatus=%02x \n", __FUNCTION__, LinkStatus);			
}



// Entry for Led control.  Triggered by timer per 50ms
void LedCtrlMain()
{	
	if(*McuIntStatus & 0x0e)	// Check if there is Tx Traffic
	{
		if( led_state != BLINK ){
			led_state = BLINK;
			ralink_gpio_led_set(LED_GPIO,1,1,4000,0,4000);
		}	
	}else{
		
		if( led_state != SOLID_ON){
			led_state = SOLID_ON;
			ralink_gpio_led_set(LED_GPIO,4000,0,1,0,4000); 
		}
	}
	
	*McuIntStatus = 0x1e; // Clear Tx and beacon Tx complete interrupt
	
	RTMPSetTimer(&LedCheckTimer, LED_CHECK_INTERVAL);
}



void ChgSignalStrengthLed()
{
	
}

void StopLed()
{
	BOOLEAN Cancelled;		
	RTMPCancelTimer(&LedCheckTimer, &Cancelled);	
	//Led Off
	ralink_gpio_led_set(LED_GPIO,0,4000,0,1,4000); 
	CheckTimerEbl=0;
	
}

void StartLed()
{
	SetLinkStatus();
}



void FlashStartLed()
{
	
	//printk(KERN_INFO "%s : LinkStatus=wuming \n", __FUNCTION__);
		
	if(CheckTimerEbl_wps==0){
		RTMPInitTimer(NULL, &LedCheckTimer_wps, LedCtrlMain_wps, NULL, TRUE);
		LedCtrlMain_wps();
		CheckTimerEbl_wps=1;
		led_state=SOLID_OFF ;
	}
}


void FlashStopLed()
{
	BOOLEAN Cancelled;		
	RTMPCancelTimer(&LedCheckTimer_wps, &Cancelled);
	//Led Off
	ralink_gpio_led_set(LED_GPIO,0,4000,0,1,4000); 
	CheckTimerEbl_wps=0;
}		

void LedCtrlMain_wps()
{
	
	if( led_state != SOLID_OFF )
	{
		ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIODATA);
		__LED_OFF(LED_GPIO);		
		*(volatile u32 *)(RALINK_REG_PIODATA) = ralink_gpio_led_value;
 	
 		//printk(KERN_INFO "%s :SOLID_OFF ralink_gpio_led_value = 0x%x  \n", __FUNCTION__, ralink_gpio_led_value);
 	
		led_state = SOLID_OFF;
	}	
	else if( led_state != SOLID_ON)
	{
		ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIODATA);
		 __LED_ON(LED_GPIO);
		*(volatile u32 *)(RALINK_REG_PIODATA) = ralink_gpio_led_value;
		
		//printk(KERN_INFO "%s :SOLID_ON ralink_gpio_led_value = 0x%x  \n", __FUNCTION__, ralink_gpio_led_value);
 			
		led_state = SOLID_ON;
		
	}	
	
	RTMPSetTimer(&LedCheckTimer_wps, LED_CHECK_INTERVAL_WPS);
}
