#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/time.h>
#include <linux/if.h>
#include "dsl.h"
#include <asm/bitops.h>

#define DRIVER "vdsl%d"
#define GIGADRIVER "giga"
#define LANDRIVER "lan%d"
#define DELAYTIME 5

#define GIGA_PORT_0_LINK_STATUS_UP 0x10

static struct net_device *vdsl_dev[9];
static int ports = 0;
//static char *type = "award5";
//MODULE_PARM(type , "s");

struct fakevdsl_private {
	struct net_device_stats stats;
	unsigned int last_time[8];
	unsigned int current_time[8];
//	void *mmio_addr; /* memory mapped IO addr */
//	unsigned long regs_len; /* length of IO or MMIO region */
};

extern int i2c_read(unsigned char chip, unsigned int addr, int alen, unsigned char *buffer, int len);
//extern int i2c_write(unsigned char chip, unsigned int addr, int alen, unsigned char *buffer, int len);

extern int smi_read(unsigned int addr, unsigned int* data);
//extern int smi_write(unsigned int addr,unsigned short data);

extern mt_ret get_phy_status(mt_uint32 portid, mt_uint8 *phy_status, mt_uint8 *conn_status);

static struct net_device_stats* fakevdsl_get_stats(struct net_device *dev)
{
#ifndef VIA_SAMSUNG
#ifdef HOPE
	unsigned int rtl8310_rx_base = 0x70d;
	unsigned int rtl8310_tx_base = 0x727;
#endif
#ifdef BOARD_A5
	unsigned int adm6996_rx_pkt_base = 0xa8;
	unsigned int adm6996_tx_pkt_base = 0xcc;
	unsigned int adm6996_rx_bytes_base = 0xba;
	unsigned int adm6996_tx_bytes_base = 0xde;
	unsigned int adm6996_colls_base = 0xf0;
	unsigned int adm6996_errors_base = 0x102;
	unsigned int value_low = 0;
	unsigned int value_high = 0;
	int shift = 0;
#endif
	mt_uint8 phy_status;
	int ret;
	unsigned int value = 0;
	int port_id = 0;
	struct timeval tv;
#endif	//VIA_SAMSUNG
	struct fakevdsl_private *tp = dev->priv;


#ifndef VIA_SAMSUNG
#ifdef HOPE
		//struct timeval tv;
		port_id = simple_strtol( &dev->name[4], NULL, 0);
		if( strcmp(dev->name , "giga") ){
			do_gettimeofday(&tv);
			tp->current_time[port_id]=tv.tv_sec;
			//printk("<1>curr_time_%d=%d\n",port_id,tp->current_time[port_id]);
			if( (tp->current_time[port_id] - tp->last_time[port_id]) >= DELAYTIME){
				ret = get_phy_status(port_id, &phy_status, NULL);
				do_gettimeofday(&tv);
				tp->last_time[port_id]=tv.tv_sec;
				//printk("<1>last_time_%d=%d\n",port_id,tp->last_time[port_id]);

				//printk("<1>ret P%d=%d\n",port_id,ret);
				//printk("<1>Status P%d=%d\n",port_id,phy_status);
				//printk("<1>OP_Status P%d=%d\n",port_id,op_status);

				if(phy_status == MT_OPSTATE_SHOWTIME){ /* Showtime */
					vdsl_dev[port_id]->flags |= IFF_UP;
					set_bit(__LINK_STATE_START, &vdsl_dev[port_id]->state);
				}else if(phy_status == MT_OPSTATE_IDLE || phy_status == MT_OPSTATE_HANDSHAKE || phy_status == MT_OPSTATE_TRAINING){ /* Idle Handshake Traning */
					//vdsl_dev[port_id]->flags &= (~IFF_UP);
					clear_bit(__LINK_STATE_START, &vdsl_dev[port_id]->state);
					vdsl_dev[port_id]->flags |= IFF_UP;
				}else{
					if(phy_status == MT_OPSTATE_DISABLE){ /* Disable */
						clear_bit(__LINK_STATE_START, &vdsl_dev[port_id]->state);
						vdsl_dev[port_id]->flags &= (~IFF_UP);
					}else{
						printk("WARRING!!:get_stats error\n");
					}
				}
			}

			i2c_read(0x54, rtl8310_rx_base+port_id , 2, (unsigned char *)&value, 4);
			tp->stats.rx_bytes = value;
			i2c_read(0x54, rtl8310_tx_base+port_id, 2, (unsigned char *)&value, 4);
			tp->stats.tx_bytes = value;
		}else{
			/* Read Giga port 0 rx bytes */
			i2c_read(0x54, 0x723, 2, (unsigned char *)&value, 4);
			tp->stats.rx_bytes = value;

			/* Read Giga port 0 tx bytes */
			i2c_read(0x54, 0x73d, 2, (unsigned char *)&value, 4);
			tp->stats.tx_bytes = value;

			/* Read Giga port 0 Link Status */
			i2c_read(0x54, 0x624, 2, (unsigned char *)&value, 4);
			value &= GIGA_PORT_0_LINK_STATUS_UP;

			if( value == GIGA_PORT_0_LINK_STATUS_UP ){
				set_bit(__LINK_STATE_START, &vdsl_dev[8]->state);
			}else if( value == 0 ){
				clear_bit(__LINK_STATE_START, &vdsl_dev[8]->state);
			}

		}
#elif defined (BOARD_A5)
		if( strcmp(dev->name , "vdsl") ){ /* dev->name is not vdsl */
			port_id = simple_strtol( &dev->name[3], NULL, 0);
			shift = 0;
			smi_read(0xa2, &value_low); /* Read switch status 0xa2 and 0xa3 */
			smi_read(0xa3, &value_high);
			value = (value_high << 16) + value_low;
						
			if(value & (0x1 << (port_id*8)) ){
				/* bit 0 is port0,bit 8 is port 1,bit 16 is port 2,bit 24 is port 3*/
				set_bit(__LINK_STATE_START, &vdsl_dev[port_id+1]->state);
				//printk("lan%d link up\n",port_id);
			}else{
				clear_bit(__LINK_STATE_START, &vdsl_dev[port_id+1]->state);
				//printk("lan%d link down\n",port_id);
			}
		}else if( !strcmp(dev->name , "vdsl") ){ /* dev->name is vdsl */
			port_id = 4;

			do_gettimeofday(&tv);
			tp->current_time[port_id]=tv.tv_sec;
			//printk("<1>curr_time_4=%d\n",tp->current_time[4]);
			if( (tp->current_time[port_id] - tp->last_time[port_id]) >= DELAYTIME){
				ret = get_phy_status(0, &phy_status, NULL);

				do_gettimeofday(&tv);
				tp->last_time[port_id]=tv.tv_sec;
				//printk("<1>last_time_0=%d\n",tp->last_time[0]);

				if(phy_status == MT_OPSTATE_SHOWTIME){ /* Showtime */
					//vdsl_dev[0]->flags |= 0x1;
					vdsl_dev[0]->flags |= IFF_UP;
					set_bit(__LINK_STATE_START, &vdsl_dev[0]->state);
				}else if(phy_status == MT_OPSTATE_IDLE || phy_status == MT_OPSTATE_HANDSHAKE || phy_status == MT_OPSTATE_TRAINING){ /* Idle Handshake Traning */
					clear_bit(__LINK_STATE_START, &vdsl_dev[0]->state);
					vdsl_dev[0]->flags |= IFF_UP;
				}else{
					if(phy_status == MT_OPSTATE_DISABLE){ /* Disable */
						clear_bit(__LINK_STATE_START, &vdsl_dev[0]->state);
						vdsl_dev[0]->flags &= (~IFF_UP);
					}else{
						printk("WARRING!!:get_stats error\n");
					}
                }
			}

			shift = 2;
		}

		do_gettimeofday(&tv);
		tp->current_time[port_id]=tv.tv_sec;
		if( (tp->current_time[port_id] - tp->last_time[port_id]) >= (DELAYTIME-1) ){
			/* Read switch rx packets counter */
			smi_read(adm6996_rx_pkt_base+(4*port_id)-shift , &value_low);
			smi_read(adm6996_rx_pkt_base+(4*port_id)-shift+1 , &value_high);
			value = (value_high << 16) + value_low;
			tp->stats.rx_packets = value;

			/* Read switch tx packets counter */
			smi_read(adm6996_tx_pkt_base+(4*port_id)-shift , &value_low);
			smi_read(adm6996_tx_pkt_base+(4*port_id)-shift+1 , &value_high);
			value = (value_high << 16) + value_low;
			tp->stats.tx_packets = value;

			/* Read switch rx bytes counter */
			smi_read(adm6996_rx_bytes_base+(4*port_id)-shift , &value_low);
			smi_read(adm6996_rx_bytes_base+(4*port_id)-shift+1 , &value_high);
			value = (value_high << 16) + value_low;
			tp->stats.rx_bytes = value;

			/* Read switch tx bytes counter */
			smi_read(adm6996_tx_bytes_base+(4*port_id)-shift , &value_low);
			smi_read(adm6996_tx_bytes_base+(4*port_id)-shift+1 , &value_high);
			value = (value_high << 16) + value_low;
			tp->stats.tx_bytes = value;

			/* Read switch collisions counter */
			smi_read(adm6996_colls_base+(4*port_id)-shift , &value_low);
			smi_read(adm6996_colls_base+(4*port_id)-shift+1 , &value_high);
			value = (value_high << 16) + value_low;
			tp->stats.collisions = value;		

			/* Read switch errors counter */
			smi_read(adm6996_errors_base+(4*port_id)-shift , &value_low);
			smi_read(adm6996_errors_base+(4*port_id)-shift+1 , &value_high);
			value = (value_high << 16) + value_low;
			tp->stats.rx_errors = value;

			do_gettimeofday(&tv);
			tp->last_time[port_id]=tv.tv_sec;
		}
#endif
#endif	//VIA_SAMSUNG

	return &(tp->stats);
}


int init_fakevdsl(void)
{
	int i=0;

//	printk("<1>Board type=%s\n",type);
#ifdef BOARD_A5
		ports = 5;
#elif defined (HOPE)
		ports = 9;
#else
		printk("<1> No type argument..using default type=award5 \n");
		ports = 5;
#endif

	/* 
	 * alloc_etherdev allocates memory for dev and dev->priv.
	 * dev->priv shall have sizeof(struct fakevdsl_private) memory
	 * allocated.
	*/
	for( i=0 ; i < ports ;i++ ){
		vdsl_dev[i] = alloc_etherdev(sizeof(struct fakevdsl_private));
	}

	for( i=0 ; i<ports ; i++){
		if(!vdsl_dev[i]) {
			printk("<1>Could not allocate vdsl%d etherdev\n",i);
			return -1;
		}
	}

#ifdef BOARD_A5
		memcpy(vdsl_dev[0]->name, "vdsl", sizeof(DRIVER)); /* Device Name */
		for(i=1 ; i<5 ; i++){
			memcpy(vdsl_dev[i]->name, LANDRIVER, sizeof(LANDRIVER)); /* Device Name */
			vdsl_dev[i]->flags |= IFF_UP; /* Lan port always up */
		}
#elif defined (HOPE)
		for( i=0 ; i<ports ; i++ ){
			memcpy(vdsl_dev[i]->name, DRIVER, sizeof(DRIVER)); /* Device Name */
		}
		memcpy(vdsl_dev[8]->name, GIGADRIVER, sizeof(GIGADRIVER)); /* Device Name */
		vdsl_dev[8]->flags |= IFF_UP; /* Giga port default up */
		//set_bit(__LINK_STATE_START, &vdsl_dev[8]->state);
#endif

	for( i=0 ; i<ports ;i++){
		vdsl_dev[i]->get_stats = fakevdsl_get_stats;
	}

	for( i=0 ; i<ports ;i++){
		if(register_netdev(vdsl_dev[i])) {
			printk("<1>Could not register netdevice vdsl%d\n",i);
		}
	}


	return 0;
}

void cleanup_fakevdsl_module(void)
{
	int i=0;

	for( i=0 ; i < ports ; i++){
		unregister_netdev(vdsl_dev[i]);
	}

	printk("<1>Unregister all devices\n");
}
