/****************************************************************************
* File:         $Workfile$
* Revision:     $Revision: 1.2 $
* Date:         $Date: 2009/06/16 08:07:25 $
*
* Overview:     
* Functions:
*
* Author:       Ian Tsao
* 
* -- Metanoia Copyright Notice --
*
* (C) Copyright 2005 Metanoia, Inc. All rights reserved.
* 
* Metanoia reserves the right to change specifications without notice. 
* Copyright 2005 Metanoia, Inc. All rights reserved. Preliminary Data-Sheet 
* indicates this product is in design and the specification may change. 
* Electrical parametrics have not been analyzed and are not specified. Do not 
* use this data sheet as a design reference document. Please contact Metanoia 
* for a current data sheet that may be used as a design reference. All 
* contents of this document are protected by copyright law and may not be 
* reproduced without the express written consent of Metanoia, Inc. Metanoia, 
* the Metanoia logo, and combinations thereof are trademarks of Metanoia, Inc. 
* Other product names used in this publication are for identification purposes 
* only and may be trademarks or registered trademarks of their respective 
* companies. The contents of this document are provided in connection with 
* Metanoia, Inc. products. Metanoia, Inc. has made best efforts to ensure that 
* the information contained herein is accurate and reliable. However, Metanoia, 
* Inc. makes no warranties, express or implied, as to the accuracy or 
* completeness of the contents of this publication and is providing this 
* publication 'as is'. Metanoia, Inc. reserves the right to make changes to 
* specifications and product descriptions at any time without notice, and to 
* discontinue or make changes to its products at any time without notice. 
* Metanoia, Inc. does not assume any liability arising out of the application 
* or use of any product or circuit, and specifically disclaims any and all 
* liability, including without limitation special, consequential, or 
* incidental damages.
* 
****************************************************************************/
#include "dsll2.h"
/*#include "mttypes.h"
#include "mterror.h"
#include "debugp.h"
#include "awardm.h"*/

void set_mapper_ports(mt_uint32 chipid, mt_uint32 ports)
{
	mt_uint32 c_map = 0;
	
	if(ports == 0)
		map[chipid] = 0;
	else
	{
		if(ports > 32)
		{
			PRINTL(DBG_LVL_MAPPER,"set_mapper_ports: Too many ports per chip %i.\n", ports);
			return;
 		}	
		while(ports)
		{
			c_map |= 1 << (--ports);
		}
 		map[chipid] = c_map;
 		PRINTL(DBG_LVL_MAPPER,"set_mapper_ports: map[%i] set to %x.\n", chipid, c_map);
	}
}

mt_uint32 get_chip_map(mt_uint32 chipid)
{
	return map[chipid];
}

mt_uint32 *get_port_map()
{
	return map;
}

mt_uint32 send_frame_mapper(mt_uint32 portid, mt_uint32 len, mt_uint8 *pframe)
{
	mt_uint8 buf[MT_MAX_BUFFERSIZE];
	mt_uint32 chipid, lportid, i;
	
	if(len+1 > MT_MAX_BUFFERSIZE)
	{
		PRINTL(DBG_LVL_MAPPER,"send_frame_mapper: Frame too long.\n");
		return MT_FRAME_TOO_LONG;
	}

	chipid = portid >> 5;
	if(chipid >= MAX_CHIPS)
	{
		PRINTL(DBG_LVL_MAPPER,"send_frame_mapper: No such chip.\n");
		return MT_NO_SUCH_CHIP;
	}
	lportid = portid & 0x1f;
	//if((map[portid >> 5] & (1 << lportid)) == 0)
	if((map[portid] & (1 << lportid)) == 0)
	{
		PRINTL(DBG_LVL_MAPPER,"send_frame_mapper: Port not available.\n");
		return MT_NO_SUCH_PORT;
	}

  	buf[0] = lportid;
	for(i = 1; i < MT_MAX_BUFFERSIZE; i++)
		buf[i] = *pframe++;

	//return send_frame_chip(chipid, len + 1, buf);
	return send_frame_chip(portid, len + 1, buf);	
}


mt_uint32 handle_rx_frame_mapper(mt_uint32 chipid, mt_uint8 *pbuf, mt_uint32 len)
{
	mt_uint8 *ptemp = pbuf;
	mt_uint32 tmp_len = len;
	PRINTL(DBG_LVL_MAPPER,"handle_rx_frame_mapper:start len=%i\n",tmp_len);
	if((tmp_len<MT_MAX_BUFFERSIZE) && ((signed int)tmp_len>0) && (debug_level & DBG_LVL_MAPPER))
	{
		PRINTL(DBG_LVL_MAPPER,"Mapper: ");
		while(tmp_len--)
			PRINTL(DBG_LVL_MAPPER,"%02x ", *ptemp++);
		PRINTL(DBG_LVL_MAPPER,"\n");
	}
	// *pbuf here is always "0", because firmware is not support to return 
	// the real port number now.
	//return handle_rx_frame_port((chipid * 32) + *pbuf, pbuf+1, len-1);
	//return handle_rx_frame_port(chipid + *pbuf, pbuf+1, len-1);
	return handle_rx_frame_port(chipid, pbuf+1, len-1);
}
