/* Configuration generator.
   Copyright (C) 2000 Kunihiro Ishiguro

This file is part of GNU Zebra.

GNU Zebra is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

GNU Zebra is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Zebra; see the file COPYING.  If not, write to the Free
Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include <zebra.h>

#include "command.h"
#include "linklist.h"
#include "memory.h"

#include "vtysh/vtysh.h"



vector configvec;

struct config
{
  /* Configuration node name. */
  char *name;

  /* Configuration string line. */
  struct list *line;

  /* Configuration can be nest. */
  struct config *config;

  /* Index of this config. */
  u_int32_t index;
};

struct list *config_top;

int
line_cmp (char *c1, char *c2)
{
  return strcmp (c1, c2);
}

void
line_del (char *line)
{
  XFREE (MTYPE_VTYSH_CONFIG_LINE, line);
}

struct config *
config_new ()
{
  struct config *config;
  config = XCALLOC (MTYPE_VTYSH_CONFIG, sizeof (struct config));
  return config;
}

int
config_cmp (struct config *c1, struct config *c2)
{
  return strcmp (c1->name, c2->name);
}

void
config_del (struct config* config)
{
  list_delete (config->line);
  if (config->name)
    XFREE (MTYPE_VTYSH_CONFIG_LINE, config->name);
  XFREE (MTYPE_VTYSH_CONFIG, config);
}

struct config *
config_get (int index, char *line)
{
  struct config *config;
  struct config *config_loop;
  struct list *master;
  struct listnode *nn;

  config = config_loop = NULL;

  master = vector_lookup_ensure (configvec, index);

  if (! master)
    {
      master = list_new ();
      master->del = (void (*) (void *))config_del;
      master->cmp = (int (*)(void *, void *)) config_cmp;
      vector_set_index (configvec, index, master);
    }
  
  LIST_LOOP (master, config_loop, nn)
    {
      if (strcmp (config_loop->name, line) == 0)
	config = config_loop;
    }

  if (! config)
    {
      config = config_new ();
      config->line = list_new ();
      config->line->del = (void (*) (void *))line_del;
      config->line->cmp = (int (*)(void *, void *)) line_cmp;
      config->name = XSTRDUP (MTYPE_VTYSH_CONFIG_LINE, line);
      config->index = index;
      listnode_add (master, config);
    }
  return config;
}

void
config_add_line (struct list *config, char *line)
{
  listnode_add (config, XSTRDUP (MTYPE_VTYSH_CONFIG_LINE, line));
}

void
config_add_line_uniq (struct list *config, char *line)
{
  struct listnode *nn;
  char *pnt;

  LIST_LOOP (config, pnt, nn)
    {
      if (strcmp (pnt, line) == 0)
	return;
    }
  listnode_add_sort (config, XSTRDUP (MTYPE_VTYSH_CONFIG_LINE, line));
}


int amit_vtysh_index = 0;

int AMIT_WRITE_CSID(char *line)
{
	int csman_fd;
    csman_fd=open_csman(NULL,0);
	int enable = 0,non_member = 999,i = 0,subnet_tmp;
	char c, *delim = " ",*pch,*pch_ip,*pch_id,*sub_line;
	struct in_addr net_ip,net_id;
	char network_ip[16],network_id[16],*network_tmp;
	struct amit_csid_ospf_network amit_csid_ospf_network[256],amit_csid_ospf_backbone;
	struct amit_csid_bgp_neighbor amit_csid_bgp_neighbor[256];
	//unsigned int area_id=0 , bbb = 16777216;
	/*if (amit_csid_init == 0)
		AMIT_GET_CSID();
	if (amit_csid_init == 1)
	{
		for (i=0;i<3;i++){
			strcpy(network_id, inet_ntoa(amit_csid_ospf_network[i].id));
			printf ("[%d]  [%s]  [%s] \n", amit_csid_ospf_network[i].enable,amit_csid_ospf_network[i].subnet,network_id);
		}
		amit_csid_init = 2;
	}*/
	
	sub_line = XMALLOC (MTYPE_TMP, strlen (line) + 1);
	sprintf (sub_line, "%s",line);
	c = sub_line[0];
	//printf ("!!!!![%s]  [%c]  [%d] \n", sub_line,c,amit_vtysh_index);
	switch (c)
	{
		case '!':
		case '#':
			amit_vtysh_index = VTYSH_INDEX_MAX;		//6
			break;
		case ' ':
			switch (amit_vtysh_index)
			{
				case VTYSH_INDEX_RIP:		//1
					break;
				case VTYSH_INDEX_OSPF:		//3
					//ospf router-id 192.168.123.254
					if 		(strncmp (sub_line, " ospf router-id", strlen (" ospf router-id")) == 0){
						delim = " ospf router-id";
						pch = strtok(sub_line,delim);
						//while (pch != NULL){
//						printf(">>%s\n",pch);
						//pch = strtok(NULL,delim);
						//write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_SUBNET+network_flag, &pch, strlen(pch)+1, CSM_W_ZERO);
						//}
					}
					//network 192.168.121.0/24 area 0.0.0.0
					else if (strncmp (sub_line, " network", strlen (" network")) == 0){
						pch 		= strtok(sub_line,delim);				// 'network'
						pch_ip 		= strtok(NULL,delim);					// "subnet"
						pch 		= strtok(NULL,delim);					// 'area'
						pch_id 		= strtok(NULL,delim);					// "area ID"
						if (strncmp (pch_id,"0.0.0.0",strlen(pch_id)) == 0)	// Check backbone ?
						{
							if (strncmp (pch_ip,amit_csid_ospf_backbone.subnet,strlen(amit_csid_ospf_backbone.subnet)) == 0)
								printf("Backbone no change(subnet:[%s])!\n",pch_ip);
							else
								write_csman(csman_fd, CSID_C_ROUTING_OSPF_BACKBONE, &pch_ip, strlen(pch_ip)+1, CSM_W_ZERO);
						}else												// Check network subnet
						{
							for (i=0;i<256;i++)
							{
//								printf("=>[%d] [%s] [%s]\n",i,amit_csid_ospf_network[i].subnet,pch_ip);	
								if (strncmp (amit_csid_ospf_network[i].subnet,pch_ip,strlen(pch_ip)) == 0)
								{
									enable = 1;
									write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_ENABLE+i, &enable, sizeof(enable), CSM_W_ZERO);
									inet_aton(pch_id,&net_id);
									write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_ID+i, &net_id, sizeof(struct in_addr), CSM_W_ZERO);
									non_member = 999 ;
									printf("ospf network[%s] no changes\n",pch_ip);	
									goto next_line;
									//break;
								}
								else if ((strlen(amit_csid_ospf_network[i].subnet) == 0)&&(non_member == 999)){
									non_member = i;
//									printf("[%d] is free!\n",non_member);	
								}
							}
							if (non_member != 999)
							{
								enable = 1;
								write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_ENABLE+non_member, &enable, sizeof(enable), CSM_W_ZERO);
								write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_SUBNET+non_member, pch_ip, strlen(pch_ip)+1, CSM_W_ZERO);
								inet_aton(pch_id,&net_id);
								write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_ID+non_member, &net_id, sizeof(struct in_addr), CSM_W_ZERO);
								printf("add ospf [network %s area %s] to %d\n",pch_ip,pch_id,non_member,strlen(pch_ip));
								non_member = 999;
							}
						}
						/*pch = strtok(pch,".");			// "area ID.1"
						area_id  = (strtol(pch,NULL,10)*16777216);
						printf(">>>>area_id1:%d\n",area_id);
						pch = strtok(NULL,".");			// "area ID.2"
						area_id += (strtol(pch,NULL,10)*65536);
						printf(">>>>area_id2:%d\n",area_id);
						pch = strtok(NULL,".");			// "area ID.3"
						area_id += (strtol(pch,NULL,10)*256);
						printf(">>>>area_id3:%d\n",area_id);
						pch = strtok(NULL,".");			// "area ID.4"
						area_id += (strtol(pch,NULL,10));
						printf(">>>>area_id4:%d\n",area_id);*/
						//write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_ID+network_flag, &pch, strlen(pch)+1, CSM_W_ZERO);

					}
					break;
				case VTYSH_INDEX_BGP:		//5
				/**/
					//bgp router-id 192.168.123.254
					if 		(strncmp (sub_line, " bgp router-id", strlen (" bgp router-id")) == 0){
						delim = " bgp router-id";
						pch = strtok(sub_line,delim);
//						printf(">>%s\n",pch);

					}
					// neighbor 192.168.13.50 remote-as 65520
					// neighbor 192.168.13.50 activate
					else if (strncmp (sub_line, " neighbor", strlen (" neighbor")) == 0){

						pch 		= strtok(sub_line,delim);				// 'neighbor'
						pch_ip 		= strtok(NULL,delim);					// "rule ip"
						pch 		= strtok(NULL,delim);					// 'remote-as'
						pch_id 		= strtok(NULL,delim);					// "rule ID"
						for (i=0;i<256;i++)
						{
							strcpy(network_ip, inet_ntoa(amit_csid_bgp_neighbor[i].ip));
							
//							printf("[%d] [%d] [%s] [%s] [%s]\n",i,non_member,network_ip,pch_ip,pch);	
							if (strncmp (network_ip,pch_ip,strlen(pch_ip)) == 0)
							{
								if (strncmp (pch,"remote-as",strlen("remote-as")) == 0)
								{
									enable = 1;
									int rule_id = strtol(pch_id,NULL,10);
									write_csman(csman_fd, CSID_C_ROUTING_BGP_RULE_ENABLE+i, &enable, sizeof(enable), CSM_W_ZERO);
									write_csman(csman_fd, CSID_C_ROUTING_BGP_RULE_ID+i, &rule_id, sizeof(rule_id), CSM_W_ZERO);
									non_member = 999 ;
									printf("bgp neighbor[%s] no changes\n",pch_ip);	
									goto next_line;
								}
								//else if (strncmp (pch,"activate",strlen("activate")) == 0)
								//{
								//	write_csman(csman_fd, CSID_C_ROUTING_BGP_RULE_ENABLE+i, &enable, sizeof(enable), CSM_W_ZERO);
								//}
								
							}
							else if ((strncmp(network_ip,"0.0.0.0",strlen("0.0.0.0")) == 0)&&(non_member == 999)){
								non_member = i;
//								printf("[%d] is free!\n",non_member);	
							}
						}
						if (non_member != 999)
						{
							enable = 1;
							write_csman(csman_fd, CSID_C_ROUTING_BGP_RULE_ENABLE+non_member, &enable, sizeof(enable), CSM_W_ZERO);
							inet_aton(pch_ip,&net_id);
							write_csman(csman_fd, CSID_C_ROUTING_BGP_RULE_IP+non_member, &net_id, sizeof(struct in_addr), CSM_W_ZERO);
							int rule_id = strtol(pch_id,NULL,10);
							write_csman(csman_fd, CSID_C_ROUTING_BGP_RULE_ID+non_member, &rule_id, sizeof(rule_id), CSM_W_ZERO);
							printf("add [network %s area %s] to %d!\n",pch_ip,pch_id,non_member);
							non_member = 999;
						}
					}
						
					//network 192.168.121.0/24
				/*	else if (strncmp (sub_line, " network", strlen (" network")) == 0){
						enable = 1;
						pch 	= strtok(sub_line,delim);				// 'network'
						pch_ip 	= strtok(NULL,delim);						// "subnet"
						read_csman (csman_fd, CSID_S_LOCAL_WAN_IP, 		&net_ip, sizeof(struct in_addr), CSM_R_ZERO);
						read_csman (csman_fd, CSID_S_ROUTING_WAN_NM_NUM, &subnet_tmp, sizeof(subnet_tmp), CSM_R_ZERO);
						strcpy(network_ip, inet_ntoa(net_ip));
						sprintf(network_tmp,"%s/%d",net_ip,subnet_tmp);
						
						if (strncmp (network_tmp,pch_ip,strlen(pch_ip)) != 0)	// Check LOCAL WAN IP ?
						{
							for (i=0;i<256;i++)
							{
								printf("[%d] [%s] [%s]\n",i,amit_csid_bgp_neighbor[i].subnet,pch_ip);	
								if (strncmp (amit_csid_bgp_neighbor[i].subnet,pch_ip,strlen(pch_ip)) == 0)
								{
									//write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_ENABLE+i, &enable, sizeof(enable), CSM_W_ZERO);
									//inet_aton(pch_id,&net_id);
									//write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_ID+i, &net_id, sizeof(struct in_addr), CSM_W_ZERO);
									non_member = 999 ;
									printf("[%s] no changes\n",pch_ip);	
									goto next_line;
								}
								else if ((strlen(amit_csid_bgp_neighbor[i].subnet) == 0)&&(non_member == 999)){
									non_member = i;
									printf("[%d] is free!\n",non_member);	
								}
							}
							if (non_member != 999)
							{
								enable = 1;
								write_csman(csman_fd, CSID_C_ROUTING_BGP_RULE_SUBNET+non_member, pch_ip, strlen(pch_ip)+1, CSM_W_ZERO);
								printf("add [network %s area %s] to %d!\n",pch_ip,pch_id,non_member);
								non_member = 999;
							}
						}
					}
				*/	
					break;
				case VTYSH_INDEX_MAX:		//6
					break;
				default:
					break;
			}	
			break;
		default:
			if (strncmp (sub_line, "router rip", strlen ("router rip")) == 0)
			{
				amit_vtysh_index = VTYSH_INDEX_RIP;		//1

			}
			else if (strncmp (sub_line, "router ospf", strlen ("router ospf")) == 0)
			{
				amit_vtysh_index = VTYSH_INDEX_OSPF;	//3
				//OSPF CSID INIT
					//network subnet area id
				amit_csid_ospf_backbone.enable = 0;
				read_csman(csman_fd, CSID_C_ROUTING_OSPF_BACKBONE, &amit_csid_ospf_backbone.subnet, sizeof(amit_csid_ospf_backbone.subnet), CSM_R_ZERO);
				amit_csid_ospf_backbone.id.s_addr = inet_addr("0.0.0.0");
				for (i=0;i<256;i++){
					amit_csid_ospf_network[i].enable = 0;
					write_csman(csman_fd, CSID_C_ROUTING_OSPF_RULE_ENABLE	+i, &amit_csid_ospf_network[i].enable, sizeof(amit_csid_ospf_network[i].enable), CSM_W_ZERO);
					read_csman (csman_fd, CSID_C_ROUTING_OSPF_RULE_SUBNET	+i, &amit_csid_ospf_network[i].subnet, sizeof(amit_csid_ospf_network[i].subnet), CSM_R_ZERO);
					read_csman (csman_fd, CSID_C_ROUTING_OSPF_RULE_ID		+i, &amit_csid_ospf_network[i].id	 , sizeof(struct in_addr), CSM_R_ZERO);
				}				
			}
			else if (strncmp (sub_line, "router bgp", strlen ("router bgp")) == 0)
			{
				amit_vtysh_index = VTYSH_INDEX_BGP;		//5
				pch_id = strtok(sub_line,"router bgp ");
//				printf(">>>>router bgp:[%s]\n",pch_id);
				unsigned int rule_id = strtol(pch_id,NULL,10);
				write_csman(csman_fd, CSID_C_ROUTING_BGP_SELFID, &rule_id, sizeof(rule_id), CSM_W_ZERO);
				//BGP CSID INIT
				for (i=0;i<256;i++){
					amit_csid_bgp_neighbor[i].enable = 0;
					write_csman(csman_fd, CSID_C_ROUTING_BGP_RULE_ENABLE	+i, &amit_csid_bgp_neighbor[i].enable, sizeof(amit_csid_bgp_neighbor[i].enable), CSM_W_ZERO);
					read_csman (csman_fd, CSID_C_ROUTING_BGP_RULE_IP		+i, &amit_csid_bgp_neighbor[i].ip, 		sizeof(struct in_addr), CSM_R_ZERO);
					read_csman (csman_fd, CSID_C_ROUTING_BGP_RULE_ID		+i, &amit_csid_bgp_neighbor[i].id	 ,sizeof(amit_csid_bgp_neighbor[i].id) , CSM_R_ZERO);
					//read_csman (csman_fd, CSID_C_ROUTING_BGP_RULE_SUBNET 	+i, &amit_csid_bgp_neighbor[i].subnet, sizeof(amit_csid_bgp_neighbor[i].subnet), CSM_R_ZERO);
				}	
				
				
			}else
				amit_vtysh_index = VTYSH_INDEX_MAX;		//6
			break;
	
	}
next_line:	
	close_csman(csman_fd);
	return 1;
}

void
vtysh_config_parse_line (char *line)
{
  char c;
  static struct config *config = NULL;

  if (! line)
    return;

  c = line[0];

  if (c == '\0')
    return;

  /* printf ("[%s]\n", line); */
	extern int amit_wrcsmn_flag;	//assigned in vtysh.c/vtysh_write_memory() 
	if (amit_wrcsmn_flag == 1)
	{	char *amit_line;
		amit_line = line;
		AMIT_WRITE_CSID(amit_line);
	}
  switch (c)
    {
    case '!':
    case '#':
      break;
    case ' ':
      /* Store line to current configuration. */
      if (config)
	{
	  if (strncmp (line, " address-family vpnv4", strlen (" address-family vpnv4")) == 0)
	    config = config_get (BGP_VPNV4_NODE, line);
	  else if (strncmp (line, " address-family ipv4", strlen (" address-family ipv4")) == 0)
	    config = config_get (BGP_IPV4_NODE, line);
	  else if (strncmp (line, " address-family ipv4 multicast", strlen (" address-family ipv4 multicast")) == 0)
	    config = config_get (BGP_IPV4M_NODE, line);
	  else if (strncmp (line, " address-family ipv6", strlen (" address-family ipv6")) == 0)
	    config = config_get (BGP_IPV6_NODE, line);
	  else if (config->index == RMAP_NODE)
	    config_add_line_uniq (config->line, line);
	  else
	    config_add_line (config->line, line);
	}
      else
	config_add_line (config_top, line);
      break;
    default:
      if (strncmp (line, "interface", strlen ("interface")) == 0)
	config = config_get (INTERFACE_NODE, line);
      else if (strncmp (line, "router rip", strlen ("router rip")) == 0)
	config = config_get (RIP_NODE, line);
      else if (strncmp (line, "router ripng", strlen ("router ripng")) == 0)
   config = config_get (RIPNG_NODE, line);
      else if (strncmp (line, "router ospf", strlen ("router ospf")) == 0)
   config = config_get (OSPF_NODE, line);
      else if (strncmp (line, "router ospf6", strlen ("router ospf6")) == 0)
   config = config_get (OSPF6_NODE, line);
      else if (strncmp (line, "router bgp", strlen ("router bgp")) == 0)
	config = config_get (BGP_NODE, line);
      else if (strncmp (line, "router", strlen ("router")) == 0)
	config = config_get (BGP_NODE, line);
      else if (strncmp (line, "route-map", strlen ("route-map")) == 0)
	config = config_get (RMAP_NODE, line);
      else if (strncmp (line, "access-list", strlen ("access-list")) == 0)
	config = config_get (ACCESS_NODE, line);
      else if (strncmp (line, "ipv6 access-list", strlen ("ipv6 access-list")) == 0)
   config = config_get (ACCESS_IPV6_NODE, line);
      else if (strncmp (line, "ip prefix-list", strlen ("ip prefix-list")) == 0)
	config = config_get (PREFIX_NODE, line);
      else if (strncmp (line, "ipv6 prefix-list", strlen ("ipv6 prefix-list")) == 0)
   config = config_get (PREFIX_IPV6_NODE, line);
      else if (strncmp (line, "ip as-path access-list", strlen ("ip as-path access-list")) == 0)
	config = config_get (AS_LIST_NODE, line);
      else if (strncmp (line, "ip community-list", strlen ("ip community-list")) == 0)
	config = config_get (COMMUNITY_LIST_NODE, line);
      else if (strncmp (line, "ip route", strlen ("ip route")) == 0)
	config = config_get (IP_NODE, line);
      else if (strncmp (line, "ipv6 route", strlen ("ipv6 route")) == 0)
   	config = config_get (IP_NODE, line);
      else if (strncmp (line, "key", strlen ("key")) == 0)
	config = config_get (KEYCHAIN_NODE, line);
      else
	{
	  if (strncmp (line, "log", strlen ("log")) == 0
	      || strncmp (line, "hostname", strlen ("hostname")) == 0
	      || strncmp (line, "password", strlen ("hostname")) == 0)
	    config_add_line_uniq (config_top, line);
	  else
	    config_add_line (config_top, line);
	  config = NULL;
	}
      break;
    }
}

void
vtysh_config_parse (char *line)
{
  char *begin;
  char *pnt;
	
  begin = pnt = line;

  while (*pnt != '\0')
    {
      if (*pnt == '\n')
	{
	  *pnt++ = '\0';
	  vtysh_config_parse_line (begin);
	  begin = pnt;
	}
      else
	{
	  pnt++;
	}
    }

}

/* Macro to check delimiter is needed between each configuration line
   or not.  */
#define NO_DELIMITER(I)  \
  ((I) == ACCESS_NODE || (I) == PREFIX_NODE || (I) == IP_NODE \
   || (I) == AS_LIST_NODE || (I) == COMMUNITY_LIST_NODE || \
   (I) == ACCESS_IPV6_NODE || (I) == PREFIX_IPV6_NODE)

/* Display configuration to file pointer.  */
void
vtysh_config_dump (FILE *fp)
{
  struct listnode *nn;
  struct listnode *nm;
  struct config *config;
  struct list *master;
  char *line;
  int i;

  LIST_LOOP (config_top, line, nn)
    {
      fprintf (fp, "%s\n", line);
      fflush (fp);
    }
  fprintf (fp, "!\n");
  fflush (fp);

  for (i = 0; i < vector_max (configvec); i++)
    if ((master = vector_slot (configvec, i)) != NULL)
      {
	LIST_LOOP (master, config, nn)
	  {
	    fprintf (fp, "%s\n", config->name);
            fflush (fp);

	    LIST_LOOP (config->line, line, nm)
	      {
		fprintf  (fp, "%s\n", line);
		fflush (fp);
	      }
	    if (! NO_DELIMITER (i))
	      {
		fprintf (fp, "!\n");
		fflush (fp);
	      }
	  }
	if (NO_DELIMITER (i))
	  {
	    fprintf (fp, "!\n");
	    fflush (fp);
	  }
      }

  for (i = 0; i < vector_max (configvec); i++)
    if ((master = vector_slot (configvec, i)) != NULL)
      {
	list_delete (master);
	vector_slot (configvec, i) = NULL;
      }
  list_delete_all_node (config_top);
}

/* Read up configuration file from file_name. */
static void
vtysh_read_file (FILE *confp)
{
  int ret;
  struct vty *vty;

  vty = vty_new ();
  vty->fd = 0;			/* stdout */
  vty->type = VTY_TERM;
  vty->node = CONFIG_NODE;
  
  vtysh_execute_no_pager ("enable");
  vtysh_execute_no_pager ("configure terminal");

  /* Execute configuration file */
  ret = vtysh_config_from_file (vty, confp);

  vtysh_execute_no_pager ("end");
  vtysh_execute_no_pager ("disable");

  vty_close (vty);

  if (ret != CMD_SUCCESS) 
    {
      switch (ret)
	{
	case CMD_ERR_AMBIGUOUS:
	  fprintf (stderr, "Ambiguous command.\n");
	  break;
	case CMD_ERR_NO_MATCH:
	  fprintf (stderr, "There is no such command.\n");
	  break;
	}
      fprintf (stderr, "Error occured during reading below line.\n%s\n", 
	       vty->buf);
      exit (1);
    }
}

/* Read up configuration file from file_name. */
void
vtysh_read_config (char *config_file,
		   char *config_current_dir,
		   char *config_default_dir)
{
  char *cwd;
  FILE *confp = NULL;
  char *fullpath;

  /* If -f flag specified. */
  if (config_file != NULL)
    {
      if (! IS_DIRECTORY_SEP (config_file[0]))
	{
	  cwd = getcwd (NULL, MAXPATHLEN);
	  fullpath = XMALLOC (MTYPE_TMP, 
			      strlen (cwd) + strlen (config_file) + 2);
	  sprintf (fullpath, "%s/%s", cwd, config_file);
	}
      else
	fullpath = config_file;

      confp = fopen (fullpath, "r");

      if (confp == NULL)
	{
	  fprintf (stderr, "can't open configuration file [%s]\n", 
		   config_file);
	  exit(1);
	}
    }
  else
    {
      /* Relative path configuration file open. */
      if (config_current_dir)
	confp = fopen (config_current_dir, "r");

      /* If there is no relative path exists, open system default file. */
      if (confp == NULL)
	{
	  confp = fopen (config_default_dir, "r");
	  if (confp == NULL)
	    {
	      fprintf (stderr, "can't open configuration file [%s]\n",
		       config_default_dir);
	      exit (1);
	    }      
	  else
	    fullpath = config_default_dir;
	}
      else
	{
	  /* Rleative path configuration file. */
	  cwd = getcwd (NULL, MAXPATHLEN);
	  fullpath = XMALLOC (MTYPE_TMP, 
			      strlen (cwd) + strlen (config_current_dir) + 2);
	  sprintf (fullpath, "%s/%s", cwd, config_current_dir);
	}  
    }  
  vtysh_read_file (confp);

  fclose (confp);

  host_config_set (fullpath);
}

void
vtysh_config_write (FILE *fp)
{
  extern struct host host;

  if (host.name)
    fprintf (fp, "hostname %s\n", host.name);
  fprintf (fp, "!\n");
}

void
vtysh_config_init ()
{
  config_top = list_new ();
  config_top->del = (void (*) (void *))line_del;
  configvec = vector_init (1);
}
